@@ -1,6 +1,4 @@
 {-# LANGUAGE OverloadedStrings   #-}
-{-# LANGUAGE QuasiQuotes         #-}
-{-# LANGUAGE TemplateHaskell     #-}
 {-# LANGUAGE ScopedTypeVariables #-}
 {-# LANGUAGE CPP                 #-}
 module Devel
@@ -16,7 +14,7 @@ import qualified Distribution.ModuleName as D
 
 import           Control.Concurrent (forkIO, threadDelay)
 import qualified Control.Exception as Ex
-import           Control.Monad (forever, when)
+import           Control.Monad (forever, when, unless)
 
 import           Data.Char (isUpper, isNumber)
 import qualified Data.List as L
@@ -29,9 +27,10 @@ import           System.FilePath (splitDirectories, dropExtension, takeExtension
 import           System.Posix.Types (EpochTime)
 import           System.PosixCompat.Files (modificationTime, getFileStatus)
 import           System.Process (createProcess, proc, terminateProcess, readProcess,
-                                           waitForProcess, rawSystem)
+                                           waitForProcess, rawSystem, runInteractiveProcess)
+import           System.IO (hClose, hIsEOF, hGetLine, stdout, stderr, hPutStrLn)
 
-import Build (recompDeps, getDeps,findHaskellFiles)
+import Build (recompDeps, getDeps)
 
 lockFile :: FilePath
 lockFile = "dist/devel-terminate"
@@ -46,9 +45,7 @@ removeLock = try_ (removeFile lockFile)
 
 devel :: Bool -> [String] -> IO ()
 devel isCabalDev passThroughArgs = do
-
     checkDevelFile
-
     writeLock
 
     putStrLn "Yesod devel server. Press ENTER to quit"
@@ -56,20 +53,20 @@ devel isCabalDev passThroughArgs = do
       cabal <- D.findPackageDesc "."
       gpd   <- D.readPackageDescription D.normal cabal
 
-      checkCabalFile gpd
+      hsSourceDirs <- checkCabalFile gpd
 
       _<- rawSystem cmd args
 
-      mainLoop
+      mainLoop hsSourceDirs
 
     _ <- getLine
     writeLock
     exitSuccess
   where
-    cmd | isCabalDev == True = "cabal-dev"
+    cmd | isCabalDev = "cabal-dev"
         | otherwise  = "cabal"
 
-    diffArgs | isCabalDev == True = [
+    diffArgs | isCabalDev = [
               "--cabal-install-arg=-fdevel" -- legacy
             , "--cabal-install-arg=-flibrary-only"
             ]
@@ -79,27 +76,28 @@ devel isCabalDev passThroughArgs = do
             ]
     args = "configure":diffArgs ++ ["--disable-library-profiling" ]
 
-    mainLoop :: IO ()
-    mainLoop = do
+    mainLoop :: [FilePath] -> IO ()
+    mainLoop hsSourceDirs = do
        ghcVer <- ghcVersion
+       when isCabalDev (rawSystemFilter cmd ["build"] >> return ())  -- cabal-dev fails with strange errors sometimes if we cabal-dev buildinfo before cabal-dev build
+       pkgArgs <- ghcPackageArgs isCabalDev ghcVer
+       let devArgs = pkgArgs ++ ["devel.hs"] ++ passThroughArgs
        forever $ do
            putStrLn "Rebuilding application..."
 
-           recompDeps
+           recompDeps hsSourceDirs
 
-           list <- getFileList
-           exit <- rawSystem cmd ["build"]
+           list <- getFileList hsSourceDirs
+           exit <- rawSystemFilter cmd ["build"]
 
            case exit of
              ExitFailure _ -> putStrLn "Build failure, pausing..."
              _ -> do
                    removeLock
-                   let pkg = pkgConfigs isCabalDev ghcVer
-                   let dev_args = pkg ++ ["devel.hs"] ++ passThroughArgs
-                   putStrLn $ "Starting development server: runghc " ++ L.intercalate " " dev_args
-                   (_,_,_,ph) <- createProcess $ proc "runghc" dev_args
+                   putStrLn $ "Starting development server: runghc " ++ L.unwords devArgs
+                   (_,_,_,ph) <- createProcess $ proc "runghc" devArgs
                    watchTid <- forkIO . try_ $ do
-                         watchForChanges list
+                         watchForChanges hsSourceDirs list
                          putStrLn "Stopping development server..."
                          writeLock
                          threadDelay 1000000
@@ -108,63 +106,52 @@ devel isCabalDev passThroughArgs = do
                    ec <- waitForProcess ph
                    putStrLn $ "Exit code: " ++ show ec
                    Ex.throwTo watchTid (userError "process finished")
-           watchForChanges list
+           watchForChanges hsSourceDirs list
 
 try_ :: forall a. IO a -> IO ()
 try_ x = (Ex.try x :: IO (Either Ex.SomeException a)) >> return ()
 
-pkgConfigs :: Bool -> String -> [String]
-pkgConfigs isCabalDev ghcVer
-  | isCabalDev = ["-package-confcabal-dev/packages-" ++ ghcVer ++ ".conf", inplacePkg]
-  | otherwise  = [inplacePkg]
-  where
-    inplacePkg = "-package-confdist/package.conf.inplace"
-
-
 type FileList = Map.Map FilePath EpochTime
 
-getFileList :: IO FileList
-getFileList = do
-    files <- findHaskellFiles "."
-    deps <- getDeps
+getFileList :: [FilePath] -> IO FileList
+getFileList hsSourceDirs = do
+    (files, deps) <- getDeps hsSourceDirs
     let files' = files ++ map fst (Map.toList deps)
     fmap Map.fromList $ flip mapM files' $ \f -> do
-        fs <- getFileStatus f
-        return (f, modificationTime fs)
+        efs <- Ex.try $ getFileStatus f
+        return $ case efs of
+            Left (_ :: Ex.SomeException) -> (f, 0)
+            Right fs -> (f, modificationTime fs)
 
-watchForChanges :: FileList -> IO ()
-watchForChanges list = do
-    newList <- getFileList
+watchForChanges :: [FilePath] ->  FileList -> IO ()
+watchForChanges hsSourceDirs list = do
+    newList <- getFileList hsSourceDirs
     if list /= newList
       then return ()
-      else threadDelay 1000000 >> watchForChanges list
+      else threadDelay 1000000 >> watchForChanges hsSourceDirs list
 
 checkDevelFile :: IO ()
 checkDevelFile = do
   e <- doesFileExist "devel.hs"
-  when (not e) $ failWith "file devel.hs not found"
+  unless e $ failWith "file devel.hs not found"
 
-checkCabalFile :: D.GenericPackageDescription -> IO ()
+checkCabalFile :: D.GenericPackageDescription -> IO [FilePath]
 checkCabalFile gpd = case D.condLibrary gpd of
     Nothing -> failWith "incorrect cabal file, no library"
     Just ct ->
-      case lookupDevelLib ct of
+      case lookupDevelLib gpd ct of
         Nothing   ->
           failWith "no development flag found in your configuration file. Expected a 'library-only' flag or the older 'devel' flag"
         Just dLib -> do
-         case (D.hsSourceDirs . D.libBuildInfo) dLib of
-           []     -> return ()
-           ["."]  -> return ()
-           _      ->
-             putStrLn $ "WARNING: yesod devel may not work correctly with " ++
-                        "custom hs-source-dirs"
-         fl <- getFileList
-         let unlisted = checkFileList fl dLib
-         when (not . null $ unlisted) $ do
-              putStrLn "WARNING: the following source files are not listed in exposed-modules or other-modules:"
-              mapM_ putStrLn unlisted
-         when (D.fromString "Application" `notElem` D.exposedModules dLib) $ do
-              putStrLn "WARNING: no exposed module Application"
+           let hsSourceDirs = D.hsSourceDirs . D.libBuildInfo $ dLib
+           fl <- getFileList hsSourceDirs
+           let unlisted = checkFileList fl dLib
+           unless (null unlisted) $ do
+                putStrLn "WARNING: the following source files are not listed in exposed-modules or other-modules:"
+                mapM_ putStrLn unlisted
+           when (D.fromString "Application" `notElem` D.exposedModules dLib) $
+                putStrLn "WARNING: no exposed module Application"
+           return hsSourceDirs
 
 failWith :: String -> IO a
 failWith msg = do
@@ -192,16 +179,44 @@ ghcVersion = fmap getNumber $ readProcess "runghc" ["--numeric-version", "0"] []
     where
       getNumber = filter (\x -> isNumber x || x == '.')
 
+ghcPackageArgs :: Bool -> String -> IO [String]
+ghcPackageArgs isCabalDev ghcVer
+  | isCabalDev = do
+      r <- readProcess "cabal-dev" ["buildopts"] []
+      let opts = L.lines r
+      return $ "-hide-all-packages" : "-no-user-package-conf" : inplacePkg : cabaldevConf : pkgid opts : depPkgIds opts
+  | otherwise = return [inplacePkg]
+      where
+        pkgid opts      = let (_,p) = head (selectOpts ["-package-name"] opts) in "-package-id" ++ p ++ "-inplace"
+        depPkgIds opts  = map (uncurry (++)) (selectOpts ["-package-id"] opts)
+        inplacePkg   = "-package-confdist/package.conf.inplace"
+        cabaldevConf = "-package-confcabal-dev/packages-" ++ ghcVer ++ ".conf"
+        selectOpts opts (x1:x2:xs)
+           | x1 `elem` opts = (x1,x2):selectOpts opts xs
+           | otherwise      = selectOpts opts (x2:xs)
+        selectOpts _ _ = []
 
-lookupDevelLib :: D.CondTree D.ConfVar c a -> Maybe a
-lookupDevelLib ct | found     = Just (D.condTreeData ct)
-                  | otherwise = Nothing
+lookupDevelLib :: D.GenericPackageDescription -> D.CondTree D.ConfVar c a -> Maybe a
+lookupDevelLib gpd ct | found     = Just (D.condTreeData ct)
+                      | otherwise = Nothing
   where
-    found = not . null . map (\(_,x,_) -> D.condTreeData x) .
-            filter isDevelLib . D.condTreeComponents  $ ct
-    isDevelLib ((D.Var (D.Flag (D.FlagName f))), _, _) = f `elem` ["library-only", "devel"]
-    isDevelLib _                                       = False
-
-
-
+    flags = map (unFlagName . D.flagName) $ D.genPackageFlags gpd
+    unFlagName (D.FlagName x) = x
+    found = any (`elem` ["library-only", "devel"]) flags
 
+-- | Acts like @rawSystem@, but filters out lines from the output that we're not interested in seeing.
+rawSystemFilter :: String -> [String] -> IO ExitCode
+rawSystemFilter command args = do
+    (inh, outh, errh, ph) <- runInteractiveProcess command args Nothing Nothing
+    hClose inh
+    let go handlein handleout = do
+            isEof <- hIsEOF handlein
+            if isEof
+                then hClose handlein
+                else do
+                    line <- hGetLine handlein
+                    unless ("Loading package " `L.isPrefixOf` line) $ hPutStrLn handleout line
+                    go handlein handleout
+    _ <- forkIO $ go outh stdout
+    _ <- forkIO $ go errh stderr
+    waitForProcess ph
